from assemblyline import odm

MSG_TYPES = {"AlerterHeartbeat"}
LOADER_CLASS = "assemblyline.odm.messages.alerter_heartbeat.AlerterMessage"


@odm.model(description="Alerter Queues")
class Queues(odm.Model):
    alert = odm.Integer(description="Number of alerts in queue")
    alert_retry = odm.Integer(description="Number of alerts in retry queue")


@odm.model(description="Alerter Metrics")
class Metrics(odm.Model):
    created = odm.Integer(description="Number of alerts created")
    error = odm.Integer(description="Number of alerts with errors")
    received = odm.Integer(description="Number of alerts received")
    updated = odm.Integer(description="Number of alerts updated")
    wait = odm.Integer(description="Number of alerts waiting for submission to complete")


@odm.model(description="Heartbeat Model for Alerter")
class Heartbeat(odm.Model):
    instances = odm.Integer(description="Number of Alerter instances")
    metrics = odm.Compound(Metrics, description="Alert metrics")
    queues = odm.Compound(Queues, description="Alert queues")


@odm.model(description="Model of Alerter Heartbeat Message")
class AlerterMessage(odm.Model):
    msg = odm.Compound(Heartbeat, description="Heartbeat message from Alerter")
    msg_loader = odm.Enum(values={LOADER_CLASS}, default=LOADER_CLASS, description="Loader class for message")
    msg_type = odm.Enum(values=MSG_TYPES, default="AlerterHeartbeat", description="Type of message")
    sender = odm.Keyword(description="Sender of message")



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.odm.models.statistics import Statistics

Classification = forge.get_classification()

DEPLOYED_STATUSES = ['DEPLOYED', 'NOISY', 'DISABLED']
DRAFT_STATUSES = ['STAGING', 'TESTING']
STALE_STATUSES = ['INVALID']

RULE_STATUSES = DEPLOYED_STATUSES + DRAFT_STATUSES + STALE_STATUSES


@odm.model(index=True, store=True)
class Signature(odm.Model):
    classification = odm.Classification(store=True, default=Classification.UNRESTRICTED)
    data = odm.Text(index=False, store=False)
    last_modified = odm.Date(default="NOW")
    name = odm.Keyword(copyto="__text__")
    order = odm.Integer(default=1, store=False)
    revision = odm.Keyword(default="1")
    signature_id = odm.Optional(odm.Keyword())
    source = odm.Keyword()
    state_change_date = odm.Optional(odm.Date(store=False))
    state_change_user = odm.Optional(odm.Keyword(store=False))
    stats = odm.Compound(Statistics, default={})
    status = odm.Enum(values=RULE_STATUSES, copyto="__text__")
    type = odm.Keyword(copyto="__text__")



from assemblyline import odm
from assemblyline.common import forge
from assemblyline.common.str_utils import StringTable

Classification = forge.get_classification()


TYPES = StringTable('TYPES', [
    ("admin", 0),
    ("user", 1),
    ("signature_manager", 2),
    ("signature_importer", 3),
    ("viewer", 4),
    ("submitter", 5),
    ("custom", 6)
])

ROLES = StringTable('ROLES', [
    ("alert_manage", 0),
    ("alert_view", 1),
    ("apikey_access", 2),
    ("bundle_download", 3),
    ("file_detail", 4),
    ("file_download", 5),
    ("file_purge", 33),
    ("heuristic_view", 6),
    ("obo_access", 7),
    ("replay_trigger", 8),
    ("safelist_view", 9),
    ("safelist_manage", 10),
    ("signature_download", 11),
    ("signature_view", 12),
    ("submission_create", 13),
    ("submission_delete", 14),
    ("submission_manage", 15),
    ("submission_view", 16),
    ("workflow_manage", 17),
    ("workflow_view", 18),
    ("administration", 19),
    ("replay_system", 20),
    ("signature_import", 21),
    ("signature_manage", 22),
    ("archive_view", 23),
    ("archive_manage", 24),
    ("archive_trigger", 25),
    ("archive_download", 26),
    ("self_manage", 27),
    ("retrohunt_view", 28),
    ("retrohunt_run", 29),
    ("external_query", 30),
    ("badlist_view", 31),
    ("badlist_manage", 32),
    ("archive_comment", 33),
    ("assistant_use", 34),
])


SCOPES = {"r", "w", "rw", "c"}
USER_TYPES = [
    TYPES.admin,               # Perform administartive task and has access to all roles
    TYPES.user,                # Normal user of the system
    TYPES.signature_manager,   # Super user that also has access to roles for managing signatures in the system
    TYPES.signature_importer,  # Has access to roles for importing signatures in the system
    TYPES.viewer,              # User that can only view the data
    TYPES.submitter,           # User that can only start submissions
    TYPES.custom,              # Has custom roles selected
]

USER_ROLES_BASIC = {
    ROLES.alert_manage,        # Modify labels, priority, status, verdict or owner of alerts
    ROLES.alert_view,          # View alerts in the system
    ROLES.archive_trigger,     # Send Submission, files and results to the archive
    ROLES.archive_view,        # View archived data in the system
    ROLES.archive_manage,      # Modify attributes of archived Submissions/Files/Results
    ROLES.archive_download,    # Download file from the archive
    ROLES.archive_comment,     # Comment on archived files
    ROLES.apikey_access,       # Allow access via API keys
    ROLES.bundle_download,     # Create bundle of a submission
    ROLES.external_query,      # Allow federated searches against external systems
    ROLES.file_detail,         # View files in the file viewer
    ROLES.file_download,       # Download files from the system
    ROLES.heuristic_view,      # View heuristics of the system
    ROLES.obo_access,          # Allow access via On Behalf Off tokens
    ROLES.replay_trigger,      # Allow submission to be replayed on another server
    ROLES.safelist_view,       # View safelist items
    ROLES.safelist_manage,     # Manage (add/delete) safelist items
    ROLES.self_manage,         # Manage currently logged in user settings
    ROLES.signature_download,  # Download signatures from the system
    ROLES.signature_view,      # View signatures
    ROLES.submission_create,   # Create a submission in the system
    ROLES.submission_delete,   # Delete submission from the system
    ROLES.submission_manage,   # Set user verdict on submissions
    ROLES.submission_view,     # View submission's results
    ROLES.workflow_manage,     # Manage (add/delete) workflows
    ROLES.workflow_view,       # View workflows
    ROLES.retrohunt_view,      # View yara searches
    ROLES.retrohunt_run,       # Run yara searches
    ROLES.badlist_view,        # View badlist items
    ROLES.badlist_manage,      # Manage (add/delete) badlist items
}

USER_ROLES = USER_ROLES_BASIC.union({
    ROLES.administration,      # Perform administrative tasks
    ROLES.file_purge,          # Purge files from the filestore
    ROLES.replay_system,       # Manage status of file/submission/alerts during the replay process
    ROLES.signature_import,    # Import signatures in the system
    ROLES.signature_manage,    # Manage signatures sources in the system
    ROLES.assistant_use,       # Use the Assemblyline Assistant
})

USER_TYPE_DEP = {
    TYPES.admin: USER_ROLES,
    TYPES.signature_importer: {
        ROLES.badlist_manage,
        ROLES.safelist_manage,
        ROLES.self_manage,
        ROLES.signature_download,
        ROLES.signature_import,
        ROLES.signature_view
    },
    TYPES.signature_manager: USER_ROLES_BASIC.union({
        ROLES.signature_manage
    }),
    TYPES.user: USER_ROLES_BASIC,
    TYPES.viewer: {
        ROLES.alert_view,
        ROLES.apikey_access,
        ROLES.badlist_view,
        ROLES.file_detail,
        ROLES.obo_access,
        ROLES.heuristic_view,
        ROLES.safelist_view,
        ROLES.self_manage,
        ROLES.signature_view,
        ROLES.submission_view,
        ROLES.workflow_view,
    },
    TYPES.submitter: {
        ROLES.apikey_access,
        ROLES.obo_access,
        ROLES.self_manage,
        ROLES.submission_create,
        ROLES.replay_trigger,
        ROLES.retrohunt_run,
    }
}

ACL_MAP = {
    "R": [
        ROLES.alert_view,
        ROLES.archive_view,
        ROLES.archive_download,
        ROLES.badlist_view,
        ROLES.bundle_download,
        ROLES.external_query,
        ROLES.file_detail,
        ROLES.file_download,
        ROLES.heuristic_view,
        ROLES.safelist_view,
        ROLES.signature_download,
        ROLES.signature_view,
        ROLES.submission_view,
        ROLES.workflow_view,
        ROLES.retrohunt_view,
    ],
    "W": [
        ROLES.alert_manage,
        ROLES.archive_trigger,
        ROLES.archive_manage,
        ROLES.badlist_manage,
        ROLES.replay_trigger,
        ROLES.safelist_manage,
        ROLES.submission_create,
        ROLES.submission_delete,
        ROLES.submission_manage,
        ROLES.retrohunt_run,
    ],
    "E": [
        ROLES.administration,
        ROLES.apikey_access,
        ROLES.file_purge,
        ROLES.obo_access,
        ROLES.replay_system,
        ROLES.self_manage,
        ROLES.signature_import,
        ROLES.signature_manage,
        ROLES.workflow_manage
    ],
    "C": []
}


def load_roles_form_acls(acls, curRoles):
    # Check if we have current roles first
    if curRoles:
        return curRoles

    # Otherwise load the roles from the api_key ACLs
    roles = set({})
    for acl in ACL_MAP.keys():
        if acl in acls:
            roles = roles.union(ACL_MAP[acl])

    # Return roles as a list
    return list(roles)


def load_roles(types, curRoles):
    # Check if we have current roles first
    if curRoles:
        return curRoles

    # Otherwise load the roles from the user type
    roles = set({})
    for user_type in USER_TYPE_DEP.keys():
        if user_type in types:
            roles = roles.union(USER_TYPE_DEP[user_type])

    # Return roles as a list
    return list(roles)


@odm.model(index=False, store=False, description="Model for API keys")
class ApiKey(odm.Model):
    acl = odm.List(odm.Enum(values=ACL_MAP.keys()), description="Access Control List for the API key")
    password = odm.Keyword(description="BCrypt hash of the password for the apikey")
    roles = odm.List(odm.Enum(values=USER_ROLES), default=[], description="List of roles tied to the API key")


@odm.model(index=False, store=False, description="Model of Apps used of OBO (On Behalf Of)")
class Apps(odm.Model):
    client_id = odm.Keyword(description="Username allowed to impersonate the current user")
    netloc = odm.Keyword(description="DNS hostname for the server")
    scope = odm.Enum(values=SCOPES, description="Scope of access for the App token")
    server = odm.Keyword(description="Name of the server that has access")
    roles = odm.List(odm.Enum(values=USER_ROLES), default=[], description="List of roles tied to the App token")


@odm.model(index=True, store=True, description="Model of User")
class User(odm.Model):
    agrees_with_tos = odm.Optional(
        odm.Date(index=False, store=False),
        description="Date the user agree with terms of service")
    api_quota = odm.Optional(odm.Integer(
        store=False, description="Maximum number of concurrent API requests (0: No Quota)"))
    api_daily_quota = odm.Optional(odm.Integer(
        store=False, description="Maximum number of API calls a user can do daily (0: No Quota)"))
    apikeys = odm.Mapping(odm.Compound(ApiKey), default={}, index=False, store=False, description="Mapping of API keys")
    apps = odm.Mapping(odm.Compound(Apps), default={}, index=False, store=False,
                       description="Applications with access to the account")
    can_impersonate = odm.Boolean(default=False, index=False, store=False,
                                  description="Allowed to query on behalf of others?")
    classification = odm.Classification(is_user_classification=True, copyto="__text__",
                                        default=Classification.UNRESTRICTED,
                                        description="Maximum classification for the user")
    dn = odm.Optional(odm.Keyword(store=False, copyto="__text__"), description="User's LDAP DN")
    email = odm.Optional(odm.Email(copyto="__text__"), description="User's email address")
    groups = odm.List(odm.UpperKeyword(), copyto="__text__", default=[],
                      description="List of groups the user submits to")
    is_active = odm.Boolean(default=True, description="Is the user active?")
    name = odm.Keyword(copyto="__text__", description="Full name of the user")
    otp_sk = odm.Optional(
        odm.Keyword(index=False, store=False),
        description="Secret key to generate one time passwords")
    password = odm.Keyword(index=False, store=False, description="BCrypt hash of the user's password")
    submission_quota = odm.Optional(odm.Integer(store=False,
                                                description="Maximum number of concurrent submissions (0: No Quota)"))
    submission_async_quota = odm.Optional(odm.Integer(
        store=False, description="Maximum number of concurrent async submission (0: No Quota)"))
    submission_daily_quota = odm.Optional(odm.Integer(
        store=False,
        description="Maximum number of submissions a user can do daily (0: No Quota)"))
    type = odm.List(odm.Enum(values=USER_TYPES), default=['user'], description="Type of user")
    roles = odm.List(odm.Enum(values=USER_ROLES), default=[], description="Default roles for user")
    security_tokens = odm.Mapping(odm.Keyword(), index=False, store=False, default={},
                                  description="Map of security tokens")
    uname = odm.Keyword(copyto="__text__", description="Username")



import enum
import os
from assemblyline.common.path import modulepath


SUBMISSION_QUEUE = 'dispatch-submission-queue'
DISPATCH_TASK_HASH = 'dispatch-active-submissions'
DISPATCH_RUNNING_TASK_HASH = 'dispatch-active-tasks'
SCALER_TIMEOUT_QUEUE = 'scaler-timeout-queue'
CONFIG_HASH = 'al-config'
POST_PROCESS_CONFIG_KEY = 'post-process-actions'


# Some pure functions for getting queue lengths (effectively for appending/prepending constants to strings)
def service_queue_name(service: str) -> str:
    """Take the name of a service, and provide the queue name to send tasks to that service."""
    return 'service-queue-' + service


def make_watcher_list_name(sid: str) -> str:
    """Get the name of the list dispatcher will pull for sending out submission events."""
    return 'dispatch-watcher-list-' + sid


def get_temporary_submission_data_name(sid: str, file_hash: str) -> str:
    """The HashMap used for tracking auxiliary processing data."""
    return '/'.join((sid, file_hash, 'temp_data'))


def get_tag_set_name(sid: str, file_hash: str) -> str:
    """The HashSet used to track the tags for an in-process file."""
    return '/'.join((sid, file_hash, 'tags'))


# A table storing information about the state of a service, expected type is ExpiringHash
# with a default ttl of None, and the ttl set per field based on the timeouts of queries
# and service operation
class ServiceStatus(enum.IntEnum):
    Idle = 0
    Running = 1


SERVICE_STATE_HASH = 'service-stasis-table'

# A null empty accepts, accepts all. A null rejects, rejects nothing
DEFAULT_SERVICE_ACCEPTS = ".*"
DEFAULT_SERVICE_REJECTS = "empty|metadata/.*"

# Queue priority values for each bucket in the ingester
PRIORITIES = {
    'low': 100,  # 0 -> 100
    'medium': 200,  # 101 -> 200
    'high': 300,
    'critical': 400,
    'user-low': 500,
    'user-medium': 1000,
    'user-high': 1500
}
MAX_PRIORITY = 2000

# The above priority values presented as a range for consistency
PRIORITY_RANGES = {}
_start = -1
for _end, _level in sorted((val, key) for key, val in PRIORITIES.items()):
    PRIORITY_RANGES[_level] = (_start + 1, _end)
    _start = _end


# Score thresholds for determining which queue priority a reingested item
# gets based on its previous score.
# eg.: item with a previous score of 99 will get 'low' priority
#      item with a previous score of 300 will get a 'high' priority
PRIORITY_THRESHOLDS = {
    'critical': 500,
    'high': 100,
}

MAGIC_RULE_PATH = os.path.join(modulepath(__name__), 'custom.magic')
YARA_RULE_PATH = os.path.join(modulepath(__name__), 'custom.yara')

MAX_INT = 2_147_483_647




import json
import logging
import tempfile
import time
from typing import Union

import requests
from assemblyline.common import forge
from assemblyline.common.identify import Identify
from assemblyline.datastore.collection import ESCollection
from assemblyline.datastore.helper import AssemblylineDatastore
from assemblyline.filestore import FileStore
from assemblyline.odm.messages.submission import Submission, SubmissionMessage
from assemblyline.odm.models.config import Config
from assemblyline.remote.datatypes import get_client
from assemblyline.remote.datatypes.queues.comms import CommsQueue
from assemblyline.remote.datatypes.queues.named import NamedQueue

try:
    from assemblyline_core.dispatching.schedules import Scheduler
    from assemblyline_core.submission_client import SubmissionClient, SubmissionException
except ImportError:
    Scheduler = SubmissionClient = SubmissionException = None

ARCHIVE_QUEUE_NAME = 'm-archive'
RETRY_MAX_BACKOFF = 5


class ArchiveManager():
    def __init__(
            self, config: Config = None, datastore: AssemblylineDatastore = None, filestore: FileStore = None,
            identify: Identify = None):
        self.log = logging.getLogger('assemblyline.archive_manager')
        self.config = config or forge.get_config()
        if self.config.datastore.archive.enabled and Scheduler:
            self.datastore = datastore or forge.get_datastore(self.config)
            self.filestore = filestore or forge.get_filestore(self.config)
            self.identify = identify or forge.get_identify(config=self.config, datastore=self.datastore, use_cache=True)
            redis_persistent = get_client(self.config.core.redis.persistent.host,
                                          self.config.core.redis.persistent.port, False)
            redis = get_client(self.config.core.redis.nonpersistent.host,
                               self.config.core.redis.nonpersistent.port, False)
            self.archive_queue: NamedQueue[dict] = NamedQueue(ARCHIVE_QUEUE_NAME, redis_persistent)
            self.scheduler = Scheduler(self.datastore, self.config, redis)
            self.submission_traffic = CommsQueue('submissions', host=redis)

    def archive_submission(self, submission, delete_after: bool = False, metadata=None, skip_hook=False,
                           use_alternate_dtl=False):
        if self.config.datastore.archive.enabled and Scheduler:
            sub_selected = self.scheduler.expand_categories(submission['params']['services']['selected'])
            min_selected = self.scheduler.expand_categories(self.config.core.archiver.minimum_required_services)

            if set(min_selected).issubset(set(sub_selected)):
                # Should we send it to a webhook ?
                if self.config.core.archiver.use_webhook and not skip_hook:
                    return {"action": "hooked", "success": self._process_hook(submission, delete_after,
                                                                              metadata, use_alternate_dtl)}

                self.archive_queue.push(('submission', submission['sid'], delete_after, metadata, use_alternate_dtl))
                return {"action": "archive", "success": True}
            else:
                params = submission['params']
                params['auto_archive'] = True
                params['delete_after_archive'] = delete_after
                params['use_archive_alternate_dtl'] = use_alternate_dtl
                params['services']['selected'] = list(set(sub_selected).union(set(min_selected)))
                if metadata and self.config.submission.metadata.archive:
                    submission['metadata'].update({k: v for k, v in metadata.items()
                                                  if k not in submission['metadata']})
                try:
                    submission_obj = Submission({
                        "files": submission["files"],
                        "metadata": submission['metadata'],
                        "params": params
                    })

                    submit_result = SubmissionClient(datastore=self.datastore, filestore=self.filestore,
                                                     config=self.config, identify=self.identify).submit(submission_obj)
                except (ValueError, KeyError) as e:
                    raise SubmissionException(f"Could not generate re-submission message: {str(e)}").with_traceback()

                self.submission_traffic.publish(SubmissionMessage({
                    'msg': submission_obj,
                    'msg_type': 'SubmissionReceived',
                    'sender': 'ui',
                }).as_primitives())

                # Update current record
                self.datastore.submission.update(submission['sid'], [(ESCollection.UPDATE_SET, 'archived', True)])

                return {"action": "resubmit", "sid": submit_result['sid'], "success": True}
        else:
            self.log.warning("Trying to archive a submission when archiving is disabled.")

    def _process_hook(self, submission: Submission, delete_after: bool = False, metadata=None, use_alternate_dtl=False):
        backoff = 0.0
        cafile = None
        hook = self.config.core.archiver.webhook

        try:
            payload = json.dumps({
                'delete_after': delete_after,
                'metadata': metadata,
                'submission': submission,
                'use_alternate_dtl': use_alternate_dtl
            })

            # Setup auth headers and other headers
            auth = None
            if hook.username and hook.password:
                auth = requests.auth.BasicAuth(login=hook.username, password=hook.password)
            headers = {head.name: head.value for head in hook.headers}
            headers.setdefault('Content-Type', 'application/json')

            # Setup ssl details
            verify: Union[None, bool, str] = None
            proxies = None
            if hook.ssl_ignore_errors:
                verify = False
            if hook.ca_cert:
                cafile = tempfile.NamedTemporaryFile()
                cafile.write(hook.ca_cert.encode())
                cafile.flush()
                verify = cafile.name
            if hook.proxy:
                proxies = {
                    "http": hook.proxy,
                    "https": hook.proxy
                }

            # Setup setup http query details
            with requests.session() as session:
                # Setup session
                session.auth = auth
                session.headers.update(headers)

                # Loop up to retry limit
                for _ in range(hook.retries):
                    # Wait before retrying, 0 first time, so we can have this before the post
                    # and not wait after the final failure
                    time.sleep(backoff)
                    backoff = min(RETRY_MAX_BACKOFF, backoff * 2) + 0.1

                    # Try posting to the webhook once. If it succeeds return and let
                    # the withs and finallys finish all the cleanup
                    try:
                        resp = session.request(hook.method, hook.uri, data=payload,
                                               verify=verify, proxies=proxies)
                        resp.raise_for_status()
                        return True
                    except Exception:
                        self.log.exception(f"Error pushing to webhook: {hook}")

        except Exception:
            self.log.exception(f"Error reading webhook configuration: {hook}")
        finally:
            if cafile is not None:
                cafile.close()

        return False



"""Sequence manipulation methods used in parsing raw datastore output."""
from __future__ import annotations
from typing import Sequence, Generator, TypeVar, overload

_T = TypeVar('_T')


@overload
def chunk(items: bytes, n: int) -> Generator[bytes, None, None]:
    ...


@overload
def chunk(items: str, n: int) -> Generator[str, None, None]:
    ...


@overload
def chunk(items: Sequence[_T], n: int) -> Generator[Sequence[_T], None, None]:
    ...


def chunk(items, n: int):
    """ Yield n-sized chunks from list.

    >>> list(chunk([1,2,3,4,5,6,7], 2))
    [[1,2], [3,4], [5,6], [7,]]
    """
    for i in range(0, len(items), n):
        yield items[i:i+n]


def chunked_list(items: Sequence[_T], n: int) -> list[Sequence[_T]]:
    """ Create a list of n-sized chunks from list.

    >>> chunked_list([1,2,3,4,5,6,7], 2)
    [[1,2], [3,4], [5,6], [7,]]
    """
    return list(chunk(items, n))



# Adjective list and noun list from: https://github.com/williexu/random_username
# Modified to allow the list to be always in memory and the choose the format

import random

ADJECTIVES = [
    "abject",
    "accomplished",
    "adept",
    "adoring",
    "adorable",
    "adventurous",
    "affectionate",
    "agreeable",
    "alert",
    "amazed",
    "amazing",
    "ambitious",
    "amiable",
    "amused",
    "approachable",
    "ardent",
    "articulate",
    "artistic",
    "attractive",
    "awed",
    "awesome",
    "beautiful",
    "blissful",
    "brainy",
    "brave",
    "bright",
    "brilliant",
    "bubbly",
    "capable",
    "charismatic",
    "charming",
    "cheerful",
    "chic",
    "clean",
    "colourful",
    "compassionate",
    "considerate",
    "content",
    "courageous",
    "courteous",
    "creative",
    "cultured",
    "curious",
    "cute",
    "dazzling",
    "dear",
    "debonair",
    "determined",
    "diligent",
    "diplomatic",
    "dynamic",
    "eager",
    "ecstatic",
    "educated",
    "efficient",
    "elegant",
    "empathic",
    "energetic",
    "engaging",
    "enthusiastic",
    "euphoric",
    "excellent",
    "excited",
    "expert",
    "fabulous",
    "faithful",
    "fantastic",
    "favourable",
    "fearless",
    "fervent",
    "fit",
    "focused",
    "fond",
    "friendly",
    "funny",
    "generous",
    "gleeful",
    "gorgeous",
    "happy",
    "helpful",
    "holistic",
    "honest",
    "humorous",
    "imaginative",
    "incredible",
    "innocent",
    "insightful",
    "intelligent",
    "inventive",
    "joyful",
    "jubilant",
    "kind",
    "knowledgeable",
    "likable",
    "loving",
    "loyal",
    "magnificent",
    "marvelous",
    "mellow",
    "merciful",
    "nice",
    "optimistic",
    "organized",
    "passionate",
    "patient",
    "peaceful",
    "perfect",
    "persistent",
    "pleased",
    "plucky",
    "polite",
    "powerful",
    "prideful",
    "productive",
    "proficient",
    "qualified",
    "reliable",
    "relieved",
    "remarkable",
    "resolved",
    "resourceful",
    "responsible",
    "sensible",
    "sincere",
    "sleek",
    "smart",
    "solid",
    "spectacular",
    "splendid",
    "spirited",
    "splendid",
    "stellar",
    "stunning",
    "stupendous",
    "sugary",
    "super",
    "superior",
    "thoughtful",
    "thrifty",
    "thrilled",
    "trusting",
    "truthful",
    "trustworthy",
    "upbeat",
    "warmhearted",
    "winged",
    "witty",
    "wonderful",
    "wondrous",
    "worldly",
    "zesty"
]

NOUNS = [
    "abalone",
    "antelope",
    "apples",
    "apricots",
    "baboon",
    "bagels",
    "basmati",
    "bass",
    "bittern",
    "boa",
    "boars",
    "bobolink",
    "buck",
    "burritos",
    "buzzard",
    "cake",
    "camel",
    "cardinal",
    "caribou",
    "caviar",
    "chamois",
    "cheese",
    "cheetah",
    "chile",
    "chough",
    "chowder",
    "clam",
    "coati",
    "cockatoo",
    "coconut",
    "cod",
    "cordial",
    "cow",
    "crackers",
    "crane",
    "cur",
    "curlew",
    "dingo",
    "dinosaur",
    "dotterel",
    "doughnut",
    "dove",
    "doves",
    "dunbird",
    "eagle",
    "eggs",
    "eland",
    "falcon",
    "ferret",
    "fish",
    "flamingo",
    "garlic",
    "gelding",
    "gnu",
    "granola",
    "hare",
    "hawk",
    "heron",
    "hoopoe",
    "hyena",
    "icecream",
    "iguana",
    "jaguar",
    "kitten",
    "lapwing",
    "lemur",
    "leopard", 
    "lion", 
    "lizard", 
    "llama", 
    "locust", 
    "lollies", 
    "macaw", 
    "mackerel", 
    "magpie", 
    "mallard", 
    "mandrill", 
    "mare", 
    "meerkat", 
    "moth", 
    "muesli", 
    "mussel", 
    "oatmeal", 
    "ocelot", 
    "oil", 
    "orange", 
    "oryx", 
    "otter", 
    "owl", 
    "paella", 
    "pear", 
    "pepper", 
    "pie", 
    "piglet", 
    "plover", 
    "polenta", 
    "ponie", 
    "porpoise", 
    "poultry", 
    "pretzels", 
    "pudding", 
    "pup", 
    "quiche", 
    "raisins", 
    "rat", 
    "relish", 
    "rhino", 
    "rice", 
    "ruffs", 
    "salami", 
    "salt", 
    "sardines", 
    "sausage", 
    "seafowl", 
    "seagull", 
    "seahorse", 
    "shads", 
    "sheep", 
    "smelt", 
    "snail", 
    "snipe", 
    "stork", 
    "swift", 
    "syrup", 
    "tacos", 
    "teal", 
    "termite", 
    "thrush", 
    "thrushe", 
    "tomatoe", 
    "tortoise", 
    "toucan", 
    "truffle", 
    "tuna", 
    "unicorn", 
    "venison", 
    "viper", 
    "wasp", 
    "weaver", 
    "whiting", 
    "widgeon", 
    "wigeon", 
    "wildfowl", 
    "zebra"
]


def random_user(digits=2, delimiter="-"):
    adjective = random.choice(ADJECTIVES)
    noun = random.choice(NOUNS)
    if digits > 0:
        num = str(random.randrange(10 ** digits))
        return f"{adjective}{delimiter}{noun}{delimiter}{num}"
    else:
        return f"{adjective}{delimiter}{noun}"
